/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "Connection.h"
#include "DrawingAreaMessagesReplies.h"
#include "MessageNames.h"
#include <WebCore/DestinationColorSpace.h>
#include <WebCore/FloatRect.h>
#include <optional>
#include <wtf/Forward.h>
#include <wtf/MachSendRight.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/text/WTFString.h>

namespace WebCore {
class FloatPoint;
class IntSize;
}

namespace WebKit {
class CallbackID;
}

namespace Messages {
namespace DrawingArea {

static inline IPC::ReceiverName messageReceiverName()
{
    return IPC::ReceiverName::DrawingArea;
}

#if USE(COORDINATED_GRAPHICS) || USE(TEXTURE_MAPPER)
class UpdateBackingStoreState {
public:
    using Arguments = std::tuple<uint64_t, bool, float, const WebCore::IntSize&, const WebCore::IntSize&>;

    static IPC::MessageName name() { return IPC::MessageName::DrawingArea_UpdateBackingStoreState; }
    static constexpr bool isSync = false;

    UpdateBackingStoreState(uint64_t backingStoreStateID, bool respondImmediately, float deviceScaleFactor, const WebCore::IntSize& size, const WebCore::IntSize& scrollOffset)
        : m_arguments(backingStoreStateID, respondImmediately, deviceScaleFactor, size, scrollOffset)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if USE(COORDINATED_GRAPHICS) || USE(TEXTURE_MAPPER)
class TargetRefreshRateDidChange {
public:
    using Arguments = std::tuple<const unsigned&>;

    static IPC::MessageName name() { return IPC::MessageName::DrawingArea_TargetRefreshRateDidChange; }
    static constexpr bool isSync = false;

    explicit TargetRefreshRateDidChange(const unsigned& rate)
        : m_arguments(rate)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class DidUpdate {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::DrawingArea_DidUpdate; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(COCOA)
class UpdateGeometry {
public:
    using Arguments = std::tuple<const WebCore::IntSize&, bool, const MachSendRight&>;

    static IPC::MessageName name() { return IPC::MessageName::DrawingArea_UpdateGeometry; }
    static constexpr bool isSync = false;

    UpdateGeometry(const WebCore::IntSize& viewSize, bool flushSynchronously, const MachSendRight& fencePort)
        : m_arguments(viewSize, flushSynchronously, fencePort)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetDeviceScaleFactor {
public:
    using Arguments = std::tuple<float>;

    static IPC::MessageName name() { return IPC::MessageName::DrawingArea_SetDeviceScaleFactor; }
    static constexpr bool isSync = false;

    explicit SetDeviceScaleFactor(float deviceScaleFactor)
        : m_arguments(deviceScaleFactor)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetColorSpace {
public:
    using Arguments = std::tuple<const std::optional<WebCore::DestinationColorSpace>&>;

    static IPC::MessageName name() { return IPC::MessageName::DrawingArea_SetColorSpace; }
    static constexpr bool isSync = false;

    explicit SetColorSpace(const std::optional<WebCore::DestinationColorSpace>& colorSpace)
        : m_arguments(colorSpace)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetViewExposedRect {
public:
    using Arguments = std::tuple<const std::optional<WebCore::FloatRect>&>;

    static IPC::MessageName name() { return IPC::MessageName::DrawingArea_SetViewExposedRect; }
    static constexpr bool isSync = false;

    explicit SetViewExposedRect(const std::optional<WebCore::FloatRect>& viewExposedRect)
        : m_arguments(viewExposedRect)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class AcceleratedAnimationDidStart {
public:
    using Arguments = std::tuple<uint64_t, const String&, const MonotonicTime&>;

    static IPC::MessageName name() { return IPC::MessageName::DrawingArea_AcceleratedAnimationDidStart; }
    static constexpr bool isSync = false;

    AcceleratedAnimationDidStart(uint64_t layerID, const String& key, const MonotonicTime& startTime)
        : m_arguments(layerID, key, startTime)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class AcceleratedAnimationDidEnd {
public:
    using Arguments = std::tuple<uint64_t, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::DrawingArea_AcceleratedAnimationDidEnd; }
    static constexpr bool isSync = false;

    AcceleratedAnimationDidEnd(uint64_t layerID, const String& key)
        : m_arguments(layerID, key)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class AddTransactionCallbackID {
public:
    using Arguments = std::tuple<const WebKit::CallbackID&>;

    static IPC::MessageName name() { return IPC::MessageName::DrawingArea_AddTransactionCallbackID; }
    static constexpr bool isSync = false;

    explicit AddTransactionCallbackID(const WebKit::CallbackID& callbackID)
        : m_arguments(callbackID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if USE(GRAPHICS_LAYER_WC)
class UpdateGeometry {
public:
    using Arguments = std::tuple<uint64_t, const WebCore::IntSize&>;

    static IPC::MessageName name() { return IPC::MessageName::DrawingArea_UpdateGeometry; }
    static constexpr bool isSync = false;

    UpdateGeometry(uint64_t backingStoreStateID, const WebCore::IntSize& viewSize)
        : m_arguments(backingStoreStateID, viewSize)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA) || PLATFORM(GTK)
class AdjustTransientZoom {
public:
    using Arguments = std::tuple<double, const WebCore::FloatPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::DrawingArea_AdjustTransientZoom; }
    static constexpr bool isSync = false;

    AdjustTransientZoom(double scale, const WebCore::FloatPoint& origin)
        : m_arguments(scale, origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA) || PLATFORM(GTK)
class CommitTransientZoom {
public:
    using Arguments = std::tuple<double, const WebCore::FloatPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::DrawingArea_CommitTransientZoom; }
    static constexpr bool isSync = false;

    CommitTransientZoom(double scale, const WebCore::FloatPoint& origin)
        : m_arguments(scale, origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

} // namespace DrawingArea
} // namespace Messages
