/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "Connection.h"
#include "FileSystemStorageError.h"
#include "MessageNames.h"
#include "NetworkStorageManagerMessagesReplies.h"
#include "SharedFileHandle.h"
#include "StorageAreaIdentifier.h"
#include "StorageAreaImplIdentifier.h"
#include "StorageAreaMapIdentifier.h"
#include "StorageNamespaceIdentifier.h"
#include <WebCore/FileSystemHandleIdentifier.h>
#include <WebCore/FileSystemSyncAccessHandleIdentifier.h>
#include <WebCore/IDBResourceIdentifier.h>
#include <WebCore/IndexedDB.h>
#include <WebCore/StorageType.h>
#include <optional>
#include <utility>
#include <wtf/Expected.h>
#include <wtf/Forward.h>
#include <wtf/HashMap.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace WebCore {
class IDBCursorInfo;
class IDBIndexInfo;
class IDBKeyData;
class IDBObjectStoreInfo;
class IDBRequestData;
class IDBResourceIdentifier;
class IDBTransactionInfo;
class IDBValue;
struct ClientOrigin;
struct IDBGetAllRecordsData;
struct IDBGetRecordData;
struct IDBIterateCursorData;
struct IDBKeyRangeData;
}

namespace Messages {
namespace NetworkStorageManager {

static inline IPC::ReceiverName messageReceiverName()
{
    return IPC::ReceiverName::NetworkStorageManager;
}

class Persisted {
public:
    using Arguments = std::tuple<const WebCore::ClientOrigin&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_Persisted; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_PersistedReply; }
    using AsyncReply = PersistedAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit Persisted(const WebCore::ClientOrigin& origin)
        : m_arguments(origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Persist {
public:
    using Arguments = std::tuple<const WebCore::ClientOrigin&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_Persist; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_PersistReply; }
    using AsyncReply = PersistAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit Persist(const WebCore::ClientOrigin& origin)
        : m_arguments(origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class FileSystemGetDirectory {
public:
    using Arguments = std::tuple<const WebCore::ClientOrigin&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_FileSystemGetDirectory; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_FileSystemGetDirectoryReply; }
    using AsyncReply = FileSystemGetDirectoryAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>>;
    explicit FileSystemGetDirectory(const WebCore::ClientOrigin& origin)
        : m_arguments(origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CloseHandle {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_CloseHandle; }
    static constexpr bool isSync = false;

    explicit CloseHandle(const WebCore::FileSystemHandleIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class IsSameEntry {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const WebCore::FileSystemHandleIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_IsSameEntry; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_IsSameEntryReply; }
    using AsyncReply = IsSameEntryAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    IsSameEntry(const WebCore::FileSystemHandleIdentifier& identifier, const WebCore::FileSystemHandleIdentifier& targetIdentifier)
        : m_arguments(identifier, targetIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetFileHandle {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const String&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_GetFileHandle; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_GetFileHandleReply; }
    using AsyncReply = GetFileHandleAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>>;
    GetFileHandle(const WebCore::FileSystemHandleIdentifier& identifier, const String& name, bool createIfNecessary)
        : m_arguments(identifier, name, createIfNecessary)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetDirectoryHandle {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const String&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_GetDirectoryHandle; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_GetDirectoryHandleReply; }
    using AsyncReply = GetDirectoryHandleAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>>;
    GetDirectoryHandle(const WebCore::FileSystemHandleIdentifier& identifier, const String& name, bool createIfNecessary)
        : m_arguments(identifier, name, createIfNecessary)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RemoveEntry {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const String&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_RemoveEntry; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_RemoveEntryReply; }
    using AsyncReply = RemoveEntryAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<std::optional<WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<std::optional<WebKit::FileSystemStorageError>>;
    RemoveEntry(const WebCore::FileSystemHandleIdentifier& identifier, const String& name, bool deleteRecursively)
        : m_arguments(identifier, name, deleteRecursively)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Resolve {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const WebCore::FileSystemHandleIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_Resolve; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<Vector<String>, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<Vector<String>, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_ResolveReply; }
    using AsyncReply = ResolveAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<Expected<Vector<String>, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<Vector<String>, WebKit::FileSystemStorageError>>;
    Resolve(const WebCore::FileSystemHandleIdentifier& identifier, const WebCore::FileSystemHandleIdentifier& targetIdentifier)
        : m_arguments(identifier, targetIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Move {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const WebCore::FileSystemHandleIdentifier&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_Move; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_MoveReply; }
    using AsyncReply = MoveAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<std::optional<WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<std::optional<WebKit::FileSystemStorageError>>;
    Move(const WebCore::FileSystemHandleIdentifier& identifier, const WebCore::FileSystemHandleIdentifier& destinationIdentifier, const String& newName)
        : m_arguments(identifier, destinationIdentifier, newName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetFile {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_GetFile; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<String, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<String, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_GetFileReply; }
    using AsyncReply = GetFileAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<Expected<String, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<String, WebKit::FileSystemStorageError>>;
    explicit GetFile(const WebCore::FileSystemHandleIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CreateSyncAccessHandle {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_CreateSyncAccessHandle; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<std::pair<WebCore::FileSystemSyncAccessHandleIdentifier, IPC::SharedFileHandle>, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<std::pair<WebCore::FileSystemSyncAccessHandleIdentifier, IPC::SharedFileHandle>, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_CreateSyncAccessHandleReply; }
    using AsyncReply = CreateSyncAccessHandleAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<Expected<std::pair<WebCore::FileSystemSyncAccessHandleIdentifier, IPC::SharedFileHandle>, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<std::pair<WebCore::FileSystemSyncAccessHandleIdentifier, IPC::SharedFileHandle>, WebKit::FileSystemStorageError>>;
    explicit CreateSyncAccessHandle(const WebCore::FileSystemHandleIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CloseSyncAccessHandle {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const WebCore::FileSystemSyncAccessHandleIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_CloseSyncAccessHandle; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_CloseSyncAccessHandleReply; }
    using AsyncReply = CloseSyncAccessHandleAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<std::optional<WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<std::optional<WebKit::FileSystemStorageError>>;
    CloseSyncAccessHandle(const WebCore::FileSystemHandleIdentifier& identifier, const WebCore::FileSystemSyncAccessHandleIdentifier& accessHandleIdentifier)
        : m_arguments(identifier, accessHandleIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetHandleNames {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_GetHandleNames; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<Vector<String>, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<Vector<String>, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_GetHandleNamesReply; }
    using AsyncReply = GetHandleNamesAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<Expected<Vector<String>, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<Vector<String>, WebKit::FileSystemStorageError>>;
    explicit GetHandleNames(const WebCore::FileSystemHandleIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetHandle {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_GetHandle; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_GetHandleReply; }
    using AsyncReply = GetHandleAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>>;
    GetHandle(const WebCore::FileSystemHandleIdentifier& identifier, const String& name)
        : m_arguments(identifier, name)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ConnectToStorageArea {
public:
    using Arguments = std::tuple<const WebCore::StorageType&, const WebKit::StorageAreaMapIdentifier&, const WebKit::StorageNamespaceIdentifier&, const WebCore::ClientOrigin&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_ConnectToStorageArea; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::StorageAreaIdentifier&&, HashMap<String, String>&&, uint64_t&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::StorageAreaIdentifier&&, HashMap<String, String>&&, uint64_t&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_ConnectToStorageAreaReply; }
    using AsyncReply = ConnectToStorageAreaAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<WebKit::StorageAreaIdentifier&, HashMap<String, String>&, uint64_t&>;
    using ReplyArguments = std::tuple<WebKit::StorageAreaIdentifier, HashMap<String, String>, uint64_t>;
    ConnectToStorageArea(const WebCore::StorageType& type, const WebKit::StorageAreaMapIdentifier& sourceIdentifier, const WebKit::StorageNamespaceIdentifier& namespaceIdentifier, const WebCore::ClientOrigin& origin)
        : m_arguments(type, sourceIdentifier, namespaceIdentifier, origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ConnectToStorageAreaSync {
public:
    using Arguments = std::tuple<const WebCore::StorageType&, const WebKit::StorageAreaMapIdentifier&, const WebKit::StorageNamespaceIdentifier&, const WebCore::ClientOrigin&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_ConnectToStorageAreaSync; }
    static constexpr bool isSync = true;

    using DelayedReply = ConnectToStorageAreaSyncDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<WebKit::StorageAreaIdentifier&, HashMap<String, String>&, uint64_t&>;
    using ReplyArguments = std::tuple<WebKit::StorageAreaIdentifier, HashMap<String, String>, uint64_t>;
    ConnectToStorageAreaSync(const WebCore::StorageType& type, const WebKit::StorageAreaMapIdentifier& sourceIdentifier, const WebKit::StorageNamespaceIdentifier& namespaceIdentifier, const WebCore::ClientOrigin& origin)
        : m_arguments(type, sourceIdentifier, namespaceIdentifier, origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CancelConnectToStorageArea {
public:
    using Arguments = std::tuple<const WebCore::StorageType&, const WebKit::StorageNamespaceIdentifier&, const WebCore::ClientOrigin&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_CancelConnectToStorageArea; }
    static constexpr bool isSync = false;

    CancelConnectToStorageArea(const WebCore::StorageType& type, const WebKit::StorageNamespaceIdentifier& namespaceIdentifier, const WebCore::ClientOrigin& origin)
        : m_arguments(type, namespaceIdentifier, origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DisconnectFromStorageArea {
public:
    using Arguments = std::tuple<const WebKit::StorageAreaIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_DisconnectFromStorageArea; }
    static constexpr bool isSync = false;

    explicit DisconnectFromStorageArea(const WebKit::StorageAreaIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CloneSessionStorageNamespace {
public:
    using Arguments = std::tuple<const WebKit::StorageNamespaceIdentifier&, const WebKit::StorageNamespaceIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_CloneSessionStorageNamespace; }
    static constexpr bool isSync = false;

    CloneSessionStorageNamespace(const WebKit::StorageNamespaceIdentifier& fromStorageNamespaceID, const WebKit::StorageNamespaceIdentifier& toStorageNamespaceID)
        : m_arguments(fromStorageNamespaceID, toStorageNamespaceID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetItem {
public:
    using Arguments = std::tuple<const WebKit::StorageAreaIdentifier&, const WebKit::StorageAreaImplIdentifier&, const String&, const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_SetItem; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&, HashMap<String, String>&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&, HashMap<String, String>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_SetItemReply; }
    using AsyncReply = SetItemAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<bool&, HashMap<String, String>&>;
    using ReplyArguments = std::tuple<bool, HashMap<String, String>>;
    SetItem(const WebKit::StorageAreaIdentifier& identifier, const WebKit::StorageAreaImplIdentifier& implIdentifier, const String& key, const String& value, const String& urlString)
        : m_arguments(identifier, implIdentifier, key, value, urlString)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RemoveItem {
public:
    using Arguments = std::tuple<const WebKit::StorageAreaIdentifier&, const WebKit::StorageAreaImplIdentifier&, const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_RemoveItem; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_RemoveItemReply; }
    using AsyncReply = RemoveItemAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    RemoveItem(const WebKit::StorageAreaIdentifier& identifier, const WebKit::StorageAreaImplIdentifier& implIdentifier, const String& key, const String& urlString)
        : m_arguments(identifier, implIdentifier, key, urlString)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Clear {
public:
    using Arguments = std::tuple<const WebKit::StorageAreaIdentifier&, const WebKit::StorageAreaImplIdentifier&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_Clear; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_ClearReply; }
    using AsyncReply = ClearAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    Clear(const WebKit::StorageAreaIdentifier& identifier, const WebKit::StorageAreaImplIdentifier& implIdentifier, const String& urlString)
        : m_arguments(identifier, implIdentifier, urlString)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class OpenDatabase {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_OpenDatabase; }
    static constexpr bool isSync = false;

    explicit OpenDatabase(const WebCore::IDBRequestData& requestData)
        : m_arguments(requestData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class OpenDBRequestCancelled {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_OpenDBRequestCancelled; }
    static constexpr bool isSync = false;

    explicit OpenDBRequestCancelled(const WebCore::IDBRequestData& requestData)
        : m_arguments(requestData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DeleteDatabase {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_DeleteDatabase; }
    static constexpr bool isSync = false;

    explicit DeleteDatabase(const WebCore::IDBRequestData& requestData)
        : m_arguments(requestData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class EstablishTransaction {
public:
    using Arguments = std::tuple<uint64_t, const WebCore::IDBTransactionInfo&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_EstablishTransaction; }
    static constexpr bool isSync = false;

    EstablishTransaction(uint64_t databaseConnectionIdentifier, const WebCore::IDBTransactionInfo& info)
        : m_arguments(databaseConnectionIdentifier, info)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DatabaseConnectionPendingClose {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_DatabaseConnectionPendingClose; }
    static constexpr bool isSync = false;

    explicit DatabaseConnectionPendingClose(uint64_t databaseConnectionIdentifier)
        : m_arguments(databaseConnectionIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DatabaseConnectionClosed {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_DatabaseConnectionClosed; }
    static constexpr bool isSync = false;

    explicit DatabaseConnectionClosed(uint64_t databaseConnectionIdentifier)
        : m_arguments(databaseConnectionIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class AbortOpenAndUpgradeNeeded {
public:
    using Arguments = std::tuple<uint64_t, const std::optional<WebCore::IDBResourceIdentifier>&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_AbortOpenAndUpgradeNeeded; }
    static constexpr bool isSync = false;

    AbortOpenAndUpgradeNeeded(uint64_t databaseConnectionIdentifier, const std::optional<WebCore::IDBResourceIdentifier>& transactionIdentifier)
        : m_arguments(databaseConnectionIdentifier, transactionIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidFireVersionChangeEvent {
public:
    using Arguments = std::tuple<uint64_t, const WebCore::IDBResourceIdentifier&, const WebCore::IndexedDB::ConnectionClosedOnBehalfOfServer&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_DidFireVersionChangeEvent; }
    static constexpr bool isSync = false;

    DidFireVersionChangeEvent(uint64_t databaseConnectionIdentifier, const WebCore::IDBResourceIdentifier& requestIdentifier, const WebCore::IndexedDB::ConnectionClosedOnBehalfOfServer& connectionClosedOnBehalfOfServer)
        : m_arguments(databaseConnectionIdentifier, requestIdentifier, connectionClosedOnBehalfOfServer)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidFinishHandlingVersionChangeTransaction {
public:
    using Arguments = std::tuple<uint64_t, const WebCore::IDBResourceIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_DidFinishHandlingVersionChangeTransaction; }
    static constexpr bool isSync = false;

    DidFinishHandlingVersionChangeTransaction(uint64_t databaseConnectionIdentifier, const WebCore::IDBResourceIdentifier& transactionIdentifier)
        : m_arguments(databaseConnectionIdentifier, transactionIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class AbortTransaction {
public:
    using Arguments = std::tuple<const WebCore::IDBResourceIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_AbortTransaction; }
    static constexpr bool isSync = false;

    explicit AbortTransaction(const WebCore::IDBResourceIdentifier& transactionIdentifier)
        : m_arguments(transactionIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CommitTransaction {
public:
    using Arguments = std::tuple<const WebCore::IDBResourceIdentifier&, uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_CommitTransaction; }
    static constexpr bool isSync = false;

    CommitTransaction(const WebCore::IDBResourceIdentifier& transactionIdentifier, uint64_t pendingRequestCount)
        : m_arguments(transactionIdentifier, pendingRequestCount)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CreateObjectStore {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBObjectStoreInfo&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_CreateObjectStore; }
    static constexpr bool isSync = false;

    CreateObjectStore(const WebCore::IDBRequestData& requestData, const WebCore::IDBObjectStoreInfo& info)
        : m_arguments(requestData, info)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DeleteObjectStore {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_DeleteObjectStore; }
    static constexpr bool isSync = false;

    DeleteObjectStore(const WebCore::IDBRequestData& requestData, const String& objectStoreName)
        : m_arguments(requestData, objectStoreName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RenameObjectStore {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, uint64_t, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_RenameObjectStore; }
    static constexpr bool isSync = false;

    RenameObjectStore(const WebCore::IDBRequestData& requestData, uint64_t objectStoreIdentifier, const String& newName)
        : m_arguments(requestData, objectStoreIdentifier, newName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ClearObjectStore {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_ClearObjectStore; }
    static constexpr bool isSync = false;

    ClearObjectStore(const WebCore::IDBRequestData& requestData, uint64_t objectStoreIdentifier)
        : m_arguments(requestData, objectStoreIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CreateIndex {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBIndexInfo&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_CreateIndex; }
    static constexpr bool isSync = false;

    CreateIndex(const WebCore::IDBRequestData& requestData, const WebCore::IDBIndexInfo& info)
        : m_arguments(requestData, info)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DeleteIndex {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, uint64_t, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_DeleteIndex; }
    static constexpr bool isSync = false;

    DeleteIndex(const WebCore::IDBRequestData& requestData, uint64_t objectStoreIdentifier, const String& indexName)
        : m_arguments(requestData, objectStoreIdentifier, indexName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RenameIndex {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, uint64_t, uint64_t, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_RenameIndex; }
    static constexpr bool isSync = false;

    RenameIndex(const WebCore::IDBRequestData& requestData, uint64_t objectStoreIdentifier, uint64_t indexIdentifier, const String& newName)
        : m_arguments(requestData, objectStoreIdentifier, indexIdentifier, newName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PutOrAdd {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBKeyData&, const WebCore::IDBValue&, const WebCore::IndexedDB::ObjectStoreOverwriteMode&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_PutOrAdd; }
    static constexpr bool isSync = false;

    PutOrAdd(const WebCore::IDBRequestData& requestData, const WebCore::IDBKeyData& key, const WebCore::IDBValue& value, const WebCore::IndexedDB::ObjectStoreOverwriteMode& overwriteMode)
        : m_arguments(requestData, key, value, overwriteMode)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetRecord {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBGetRecordData&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_GetRecord; }
    static constexpr bool isSync = false;

    GetRecord(const WebCore::IDBRequestData& requestData, const WebCore::IDBGetRecordData& getRecordData)
        : m_arguments(requestData, getRecordData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetAllRecords {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBGetAllRecordsData&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_GetAllRecords; }
    static constexpr bool isSync = false;

    GetAllRecords(const WebCore::IDBRequestData& requestData, const WebCore::IDBGetAllRecordsData& getAllRecordsData)
        : m_arguments(requestData, getAllRecordsData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetCount {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBKeyRangeData&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_GetCount; }
    static constexpr bool isSync = false;

    GetCount(const WebCore::IDBRequestData& requestData, const WebCore::IDBKeyRangeData& range)
        : m_arguments(requestData, range)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DeleteRecord {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBKeyRangeData&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_DeleteRecord; }
    static constexpr bool isSync = false;

    DeleteRecord(const WebCore::IDBRequestData& requestData, const WebCore::IDBKeyRangeData& range)
        : m_arguments(requestData, range)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class OpenCursor {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBCursorInfo&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_OpenCursor; }
    static constexpr bool isSync = false;

    OpenCursor(const WebCore::IDBRequestData& requestData, const WebCore::IDBCursorInfo& info)
        : m_arguments(requestData, info)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class IterateCursor {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBIterateCursorData&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_IterateCursor; }
    static constexpr bool isSync = false;

    IterateCursor(const WebCore::IDBRequestData& requestData, const WebCore::IDBIterateCursorData& data)
        : m_arguments(requestData, data)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetAllDatabaseNamesAndVersions {
public:
    using Arguments = std::tuple<const WebCore::IDBResourceIdentifier&, const WebCore::ClientOrigin&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_GetAllDatabaseNamesAndVersions; }
    static constexpr bool isSync = false;

    GetAllDatabaseNamesAndVersions(const WebCore::IDBResourceIdentifier& requestIdentifier, const WebCore::ClientOrigin& origin)
        : m_arguments(requestIdentifier, origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

} // namespace NetworkStorageManager
} // namespace Messages
