/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(SERVICE_WORKER)

#include "ArgumentCoders.h"
#include "Connection.h"
#include "MessageNames.h"
#include "SharedMemory.h"
#include "WebSWClientConnectionMessagesReplies.h"
#include <WebCore/FetchOptions.h>
#include <WebCore/ProcessQualified.h>
#include <WebCore/ScriptExecutionContextIdentifier.h>
#include <WebCore/ServiceWorkerClientData.h>
#include <WebCore/ServiceWorkerData.h>
#include <WebCore/ServiceWorkerIdentifier.h>
#include <WebCore/ServiceWorkerTypes.h>
#include <optional>
#include <wtf/Forward.h>
#include <wtf/HashSet.h>
#include <wtf/ObjectIdentifier.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/text/WTFString.h>

namespace WebCore {
class ServiceWorkerRegistrationKey;
enum class ShouldNotifyWhenResolved : bool;
enum class ServiceWorkerRegistrationState : uint8_t;
enum class ServiceWorkerState : uint8_t;
enum class ServiceWorkerUpdateViaCache : uint8_t;
struct ExceptionData;
struct MessageWithMessagePorts;
struct ServiceWorkerData;
struct ServiceWorkerRegistrationData;
}

namespace Messages {
namespace WebSWClientConnection {

static inline IPC::ReceiverName messageReceiverName()
{
    return IPC::ReceiverName::WebSWClientConnection;
}

class JobRejectedInServer {
public:
    using Arguments = std::tuple<const WebCore::ServiceWorkerJobIdentifier&, const WebCore::ExceptionData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebSWClientConnection_JobRejectedInServer; }
    static constexpr bool isSync = false;

    JobRejectedInServer(const WebCore::ServiceWorkerJobIdentifier& jobDataIdentifier, const WebCore::ExceptionData& exception)
        : m_arguments(jobDataIdentifier, exception)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RegistrationJobResolvedInServer {
public:
    using Arguments = std::tuple<const WebCore::ServiceWorkerJobIdentifier&, const WebCore::ServiceWorkerRegistrationData&, WebCore::ShouldNotifyWhenResolved>;

    static IPC::MessageName name() { return IPC::MessageName::WebSWClientConnection_RegistrationJobResolvedInServer; }
    static constexpr bool isSync = false;

    RegistrationJobResolvedInServer(const WebCore::ServiceWorkerJobIdentifier& jobDataIdentifier, const WebCore::ServiceWorkerRegistrationData& registration, WebCore::ShouldNotifyWhenResolved shouldNotifyWhenResolved)
        : m_arguments(jobDataIdentifier, registration, shouldNotifyWhenResolved)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StartScriptFetchForServer {
public:
    using Arguments = std::tuple<const WebCore::ServiceWorkerJobIdentifier&, const WebCore::ServiceWorkerRegistrationKey&, const WebCore::FetchOptions::Cache&>;

    static IPC::MessageName name() { return IPC::MessageName::WebSWClientConnection_StartScriptFetchForServer; }
    static constexpr bool isSync = false;

    StartScriptFetchForServer(const WebCore::ServiceWorkerJobIdentifier& jobDataIdentifier, const WebCore::ServiceWorkerRegistrationKey& registrationKey, const WebCore::FetchOptions::Cache& cachePolicy)
        : m_arguments(jobDataIdentifier, registrationKey, cachePolicy)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class UpdateRegistrationState {
public:
    using Arguments = std::tuple<const WebCore::ServiceWorkerRegistrationIdentifier&, WebCore::ServiceWorkerRegistrationState, const std::optional<WebCore::ServiceWorkerData>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebSWClientConnection_UpdateRegistrationState; }
    static constexpr bool isSync = false;

    UpdateRegistrationState(const WebCore::ServiceWorkerRegistrationIdentifier& identifier, WebCore::ServiceWorkerRegistrationState state, const std::optional<WebCore::ServiceWorkerData>& serviceWorkerIdentifier)
        : m_arguments(identifier, state, serviceWorkerIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class UpdateWorkerState {
public:
    using Arguments = std::tuple<const WebCore::ServiceWorkerIdentifier&, WebCore::ServiceWorkerState>;

    static IPC::MessageName name() { return IPC::MessageName::WebSWClientConnection_UpdateWorkerState; }
    static constexpr bool isSync = false;

    UpdateWorkerState(const WebCore::ServiceWorkerIdentifier& serviceWorkerIdentifier, WebCore::ServiceWorkerState state)
        : m_arguments(serviceWorkerIdentifier, state)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class FireUpdateFoundEvent {
public:
    using Arguments = std::tuple<const WebCore::ServiceWorkerRegistrationIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebSWClientConnection_FireUpdateFoundEvent; }
    static constexpr bool isSync = false;

    explicit FireUpdateFoundEvent(const WebCore::ServiceWorkerRegistrationIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetRegistrationLastUpdateTime {
public:
    using Arguments = std::tuple<const WebCore::ServiceWorkerRegistrationIdentifier&, const WallTime&>;

    static IPC::MessageName name() { return IPC::MessageName::WebSWClientConnection_SetRegistrationLastUpdateTime; }
    static constexpr bool isSync = false;

    SetRegistrationLastUpdateTime(const WebCore::ServiceWorkerRegistrationIdentifier& identifier, const WallTime& lastUpdateTime)
        : m_arguments(identifier, lastUpdateTime)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetRegistrationUpdateViaCache {
public:
    using Arguments = std::tuple<const WebCore::ServiceWorkerRegistrationIdentifier&, WebCore::ServiceWorkerUpdateViaCache>;

    static IPC::MessageName name() { return IPC::MessageName::WebSWClientConnection_SetRegistrationUpdateViaCache; }
    static constexpr bool isSync = false;

    SetRegistrationUpdateViaCache(const WebCore::ServiceWorkerRegistrationIdentifier& identifier, WebCore::ServiceWorkerUpdateViaCache updateViaCache)
        : m_arguments(identifier, updateViaCache)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class NotifyClientsOfControllerChange {
public:
    using Arguments = std::tuple<const HashSet<WebCore::ScriptExecutionContextIdentifier>&, const WebCore::ServiceWorkerData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebSWClientConnection_NotifyClientsOfControllerChange; }
    static constexpr bool isSync = false;

    NotifyClientsOfControllerChange(const HashSet<WebCore::ScriptExecutionContextIdentifier>& contextIdentifiers, const WebCore::ServiceWorkerData& newController)
        : m_arguments(contextIdentifiers, newController)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetSWOriginTableIsImported {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebSWClientConnection_SetSWOriginTableIsImported; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetSWOriginTableSharedMemory {
public:
    using Arguments = std::tuple<const WebKit::SharedMemory::Handle&>;

    static IPC::MessageName name() { return IPC::MessageName::WebSWClientConnection_SetSWOriginTableSharedMemory; }
    static constexpr bool isSync = false;

    explicit SetSWOriginTableSharedMemory(const WebKit::SharedMemory::Handle& handle)
        : m_arguments(handle)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PostMessageToServiceWorkerClient {
public:
    using Arguments = std::tuple<const WebCore::ScriptExecutionContextIdentifier&, const WebCore::MessageWithMessagePorts&, const WebCore::ServiceWorkerData&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebSWClientConnection_PostMessageToServiceWorkerClient; }
    static constexpr bool isSync = false;

    PostMessageToServiceWorkerClient(const WebCore::ScriptExecutionContextIdentifier& destinationContextIdentifier, const WebCore::MessageWithMessagePorts& message, const WebCore::ServiceWorkerData& source, const String& sourceOrigin)
        : m_arguments(destinationContextIdentifier, message, source, sourceOrigin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetServiceWorkerClientIsControlled {
public:
    using Arguments = std::tuple<const WebCore::ScriptExecutionContextIdentifier&, const WebCore::ServiceWorkerRegistrationData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebSWClientConnection_SetServiceWorkerClientIsControlled; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebSWClientConnection_SetServiceWorkerClientIsControlledReply; }
    using AsyncReply = SetServiceWorkerClientIsControlledAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    SetServiceWorkerClientIsControlled(const WebCore::ScriptExecutionContextIdentifier& workerIdentifier, const WebCore::ServiceWorkerRegistrationData& data)
        : m_arguments(workerIdentifier, data)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class FocusServiceWorkerClient {
public:
    using Arguments = std::tuple<const WebCore::ScriptExecutionContextIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebSWClientConnection_FocusServiceWorkerClient; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(std::optional<WebCore::ServiceWorkerClientData>&&)>&&);
    static void cancelReply(CompletionHandler<void(std::optional<WebCore::ServiceWorkerClientData>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebSWClientConnection_FocusServiceWorkerClientReply; }
    using AsyncReply = FocusServiceWorkerClientAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    using Reply = std::tuple<std::optional<WebCore::ServiceWorkerClientData>&>;
    using ReplyArguments = std::tuple<std::optional<WebCore::ServiceWorkerClientData>>;
    explicit FocusServiceWorkerClient(const WebCore::ScriptExecutionContextIdentifier& clientIdentifier)
        : m_arguments(clientIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

} // namespace WebSWClientConnection
} // namespace Messages

#endif // ENABLE(SERVICE_WORKER)
